//name of file: classes.cpp

//instructions for the preprocessor
#include<stdlib.h>
#include<fstream.h>
#include<iomanip.h>
//#include<strstream.h>
#include<sstream>      //13.04.05
#include<math.h>
#include<time.h>

#define sgn(x) ((x)>0 ? 1 : ((x)==0 ? 0 : -1))
#define abs(x) ((x)>=0 ? (x) : (-x))
#define min(x,y) (((x)>=(y)) ? (y) : (x) )
#define max(x,y) (((x)>=(y)) ? (x) : (y) )
#define FALSE  0
#define TRUE   1
#define EXTERN 1
#define INTERN 0
#define Pi  3.14159


//the first part of this file contains some general stuff, which is
//not specific to the meshing-generator

//randombit yields a randomly chosen bit i.e. either 0 or 1.
//initialisation of the random-generator: srand(time(NULL));

bool randombit(){
  unsigned long r;
  r=rand();
  if(r<RAND_MAX/2) return 0;
  else return 1;
}

///////////////////////////////////////////////////////////////////

template<class T>
void swap(T& a,T& b){
  T c;
  c=a; a=b; b=c;
}

template<class T>
int* select_sort(T *array, int N){
  T least;
  int* tmp;
  int* perm ;//= new int[N];
  int k;

  tmp  = (int*) malloc(sizeof(int)*N);
  perm = (int*) malloc(sizeof(int)*N);
  for(int i=0; i<N; i++) tmp[i]=i;
  for(int i=0; i<N; i++){
    least = array[i]; k=i;
    for(int j=i; j<N; j++)
      if(array[j] < least){ least = array[j]; k=j;}
    swap(array[i],array[k]);
    swap(tmp[i],tmp[k]);
  }
  //if i is the new index of an array-element (after sorting)
  //then tmp[i] yields the old index (before sorting).
  //To get the new index knowing the old one the permutation tmp is
  //inverted and saved in perm !
  for(int i=0; i<N; i++) perm[tmp[i]] = i;
  free(tmp); return perm;
}//select_sort

///////////////////////////////////////////////////////////////////
//definition of an element of a doubly connected list
//A doubly connected ring-list is used for the construction of the
//triangulation - the data member is called "edgep" because it
//represents a pointer of type CEdge*

template<class T>
class  DListElement{
  public:
    T edgep;                 //data-member
    DListElement *next;      //pointer to the next/following element
    DListElement *previous;  //pointer to the precedent element
};

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
//general definition of a stack:  last in - first out  (LIFO)

template<class T>
class StackElement{
  public:
    StackElement* prev;
    T data;

    void add(T data, StackElement* &top){
      StackElement* newtop = new StackElement;
      newtop->data = data;
      newtop->prev = top;
      top = newtop;
    };

    T extract(StackElement* &top){
      T data = top->data;
      StackElement* tmp = top->prev;
      delete top; top = tmp;
      return data;
    };
};


template<class T>
class Stack{
  public:
    Stack(){counter = 0; top=NULL;};
   ~Stack(){while(counter>0){extract(top); counter--;}};
    int cardinal(){return counter;};
    void push(T data){counter++; add(data, top);};
    T pop(){
      if(counter==0){cerr<<"Error: You can't pop an element from a void stack"<<endl; exit(1);}
      else{counter--; return extract(top);}
    };
  private:
  int counter;
  StackElement<T>* top;
};

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
// general implementation of a queue: first in first out  (FIFO)

template<class T>
class QueueElement{
  public:
    QueueElement* prev;
    T data;

    friend T take<>(QueueElement<T>* &head);
    friend void add<>(T data, QueueElement<T>* &tail);
    friend void init<>(T data, QueueElement<T>* &tail, QueueElement<T>* &head);
};

template<class T>
T take(QueueElement<T>* &head){
  T data = head->data;
  QueueElement<T>* tmp = head->prev;
  delete head; head = tmp;
  return data;
}

template<class T>
void add(T data, QueueElement<T>* &tail){
  QueueElement<T>* newtail = new QueueElement<T>;
  newtail->data = data;
  newtail->prev = NULL;
  tail->prev = newtail;
  tail = newtail;
}

template<class T>
void init(T data, QueueElement<T>* &tail, QueueElement<T>* &head){
  QueueElement<T>* first = new QueueElement<T>;
  first->data = data;
  first->prev = NULL;
  tail = first;
  head = first;
}

template<class T>
class Queue{
  public:
  Queue(){counter = 0;};
 ~Queue(){while(counter > 0){take(head); counter--;}};
  int  cardinal(){return counter;};
  void push(T data){
    if     (counter >0){add(data,tail); counter++;}
    else if(counter==0){init(data,tail,head); counter++;}
  };
  T pop(){
     if(counter==0){cerr<<"You can't pop an element from a void queue !"<<endl; exit(1);}
     if(counter >0){if(counter==1)tail = NULL; //
                    counter--; return take(head);}
  };
  private:
  int counter;
  QueueElement<T> *tail;
  QueueElement<T> *head;
};

////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////

//declaration of class K2 modelling vectors and points in the plane

typedef  float K;

class K2{
  public:
    K2(K a=0,K b=0):x(a),y(b){};            //default constructor
    K2(const K2& v){x=v.x; y=v.y;};         //copy constructor
   ~K2(){};                                 //default destructor

    K2& operator = (const K2&);
    K2& operator +=(const K2&);
    K2& operator -=(const K2&);
    K2& operator *=(const K&);
    K2  operator+ (const K2&) const;
    K2  operator- (const K2&) const;
    K   operator* (const K2&) const;
    inline K&  operator [](int i);
    bool operator <(const K2&) const;

    friend K Angle(const K2&, const K2&);
    friend K Det  (const K2&, const K2&);
    friend K Norm (const K2&);
    friend ostream& operator<< (ostream&, const K2&);
    friend fstream& operator<< (fstream&, const K2&);
    friend fstream& operator>> (fstream&, K2&);
  protected:
    K x,y;
};

// definition of methods and friend functions of class K2
K2& K2::operator= (const K2& v){x=v.x; y=v.y;   return *this;}
K2& K2::operator+=(const K2& v){x+=v.x; y+=v.y; return *this;}
K2& K2::operator-=(const K2& v){x-=v.x; y-=v.y;  return *this;}
K2  K2::operator+ (const K2& w)const {K2 u; u.x=x+w.x; u.y=y+w.y; return u;}
K2  K2::operator- (const K2& w)const {K2 u; u.x=x-w.x; u.y=y-w.y; return u;}
K  K2::operator* (const K2& w)const {return x*w.x + y*w.y;}
inline K& K2::operator[] (int i)
{if(i==1) return x; else if(i==2) return y; else exit(1);}

bool K2::operator <(const K2& v) const{
  K a = Norm(*this);
  K b = Norm(v);
  if(a < b) return TRUE;
  else if(a==b && abs(x) < abs(v.x)) return TRUE;
  else return FALSE;
}

K Det  (const K2& v, const K2& w){return v.x*w.y - v.y*w.x;}
K Norm (const K2& v){return sqrt(v.x*v.x + v.y*v.y);}

K Angle(K2& v, K2& w){
//calculates the angle phi between two vectors
//By convention we set: -Pi < phi <= Pi
//furthermore, phi is negative, if the vectors do not form a
//righthand system, i.e. if we turn v counterclockwise to get
//parallel with w we must turn about an angle larger than Pi.
  if(Norm(v)==0 || Norm(w)==0) return 0;
  else{
    K cosinus = (v*w)/(Norm(v)*Norm(w));
    K sinus   = Det(v,w)/(Norm(v)*Norm(w));
    if(cosinus != 0){
      K tangens = sinus/cosinus;
      K phi = atan(tangens);
      if     (sinus == 0){
        if     (cosinus<0)            return  0 ;
        else if(cosinus>0)            return  Pi;  }
      else if(tangens < 0){
        if     (cosinus<0 && sinus>0) return phi+Pi;
        else if(cosinus>0 && sinus<0) return phi;  }
      else if(tangens > 0){
        if     (cosinus<0 && sinus<0) return phi-Pi;
        else if(cosinus>0 && sinus>0) return phi;  }
    }else{// cosinus ==0
      if(sinus > 0) return  Pi/2;
      else          return -Pi/2;
    } // else
  }// first else
}

ostream& operator<< (ostream& output, const K2& v){
  output.setf(ios::right | ios::showpos | ios::fixed);
  output.precision(4);
  output.width(11);
  output << "r = "<<Norm(v);
  output.width(11);
  output << "x = "<<v.x;
  output.width(11);
  output << "y = "<<v.y << endl;
}

fstream& operator<< (fstream& finout, const K2& v){
  finout.setf(ios::right | ios::showpos | ios::fixed);
  finout.precision(4);
  finout.width(11);
  //finout << v.x;     13.04.05
  finout.width(11);
  //finout << v.y << endl;  13.04.05
}

fstream& operator>> (fstream& finout,  K2& v){
  finout >> v.x >> v.y;
}
// end class K2
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
//definition of class CVertex

class CEdge; //predeclaration of class CEdge necessary for defining data-member "CEdge *(*e)"

class CVertex: public K2 {
  public:
  int ne;      //number of edges containing the vertex as head or foot
  CEdge *(*e); //array of pointer to edge
  int bflag;   //bflag is -1 if the vertex is not lying on the boundary
               //otherwise it corresponds to the boundary-index (0<= bflag <bnum)

  CVertex(){ne=0; e=NULL;};
 ~CVertex(){};
  void operator = (const CVertex& W) {x=W.x; y=W.y; ne=W.ne; e=W.e; bflag = W.bflag;};
  //operator K2() const{K2 vector(x,y); return vector;}
  void operator += (K2 v){x+=v[1]; y+=v[2];}; //x+=v.x
  void operator -= (K2 v){x-=v[1]; y-=v[2];};
};

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

class CEdge{
  public:
    CEdge(){foot=-1; head=-1; left=-1; right=-1;
            fln=0; frn=0; hln=0; hrn=0;
            bflag=FALSE; cflag=-1;};
   ~CEdge(){};
    int foot;          //number of foot-vertex
    int head;          //number of head-vertex
    bool status;       //possible values: INTERN=0 EXTERN=1
    int left;          //number of vertex to the left
    int right;         //number of vertex to the right
    CEdge *fln;        //pointer to left-neighbor-edge at foot-vertex
    CEdge *frn;        //pointer to right-neighbor-edge at foot-vertex
    CEdge *hln;        //pointer to left-neighbor-edge at head-vertex
    CEdge *hrn;        //pointer to right-neighbor-edge at head-vertex
    bool bflag;        //boundary-flag
    int  cflag;        //color-flag


    void reverse(void); //alternative CEdge& reverse(void);
    inline int& relleft(int);     //relative left vertex
    inline int& relright(int);
    inline CEdge*& relhln(int);   //relative head-left-neighbor
    inline CEdge*& relhrn(int);   //relative head-right-neighbor
    inline CEdge*& relfln(int);
    inline CEdge*& relfrn(int);
    inline int oppend(int);     //opposite end
    inline int& oppvertex(int); //opposite vertex
    inline CEdge*& oppneighbor(CEdge*);
    friend ostream& operator<< (ostream&, const CEdge&);
};


void CEdge::reverse(void){
  int tmp;
  CEdge* temp1;
  CEdge* temp2;
  tmp = foot; foot = head;  head  = tmp;
  tmp = left; left = right; right = tmp;
  temp1 = fln;  fln = hrn; hrn = temp1;
  temp2 = hln;  hln = frn; frn = temp2;
}


inline int& CEdge::relleft(int vertexindex){
   if(vertexindex==foot) return left;
   else if(vertexindex==head) return right;
   else cout<<"relleft: index error" << endl;
}

inline int& CEdge::relright(int vertexindex){
   if(vertexindex==foot) return right;
   else if(vertexindex==head) return left;
   else cout<<"relright: index error" << endl;
}

inline CEdge*& CEdge::relhln(int vertexindex){
   if(vertexindex==foot) return hln;
   else if(vertexindex==head) return frn;
   else cout<<"relhln: index error"<< endl;
}

inline CEdge*& CEdge::relhrn(int vertexindex){
   if(vertexindex==foot) return hrn;
   else if(vertexindex==head) return fln;
   else cout<<"relhrn: index error"<< endl;
}

inline CEdge*& CEdge::relfln(int vertexindex){
   if(vertexindex==foot) return fln;
   else if(vertexindex==head) return hrn;
   else cout<<"relfln: index error"<< endl;
}

inline CEdge*& CEdge::relfrn(int vertexindex){
   if(vertexindex==foot) return frn;
   else if(vertexindex==head) return hln;
   else cout<<"relfrn: index error"<< endl;
}

inline int CEdge::oppend(int vertexindex){
   if(vertexindex == foot) return head;
   else if(vertexindex == head) return foot;
   else cerr<<"oppend: index error"<<endl;
}

inline int& CEdge::oppvertex(int vertexindex){
   if(vertexindex == left) return right;
   else if(vertexindex == right) return left;
   else cerr<<"oppvertex: index error"<<endl;
}

inline CEdge*& CEdge::oppneighbor(CEdge* neighbor){
   if(neighbor==fln)      return frn;
   else if(neighbor==frn) return fln;
   else if(neighbor==hln) return hrn;
   else if(neighbor==hrn) return hln;
   else cout<<"oppneighbor: error"<<endl;
}


ostream& operator<< (ostream& output, const CEdge& edge){
  output.setf(ios::right | ios::fixed);
  output << "   foot = ";
  output.width(4);
  output << edge.foot;
  output << "   head = ";
  output.width(4);
  output << edge.head ;
  output << "   bflag = "<<edge.bflag;
  output << "   cflag  ="<<edge.cflag;
  output << endl;
}
//end class CEdge
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

class CBoundary{
  public:
    int first;  //contains the index (before sorting) of the first boundary vertex
    int last;   //contains the index (before sorting) of the last boundary vertex
};

///////////////////////////////////////////////////////////////////

class CTriangle{
  public:
    int v[3];  //contains the indices of the three vertices
    friend ostream& operator<< (ostream&, const CTriangle&);
    friend ofstream& operator<< (ofstream&, const CTriangle&);
};

ostream& operator<< (ostream& output, const CTriangle& triangle){
  //output<<"   1st vertex ";
  output.width(10);
  output<< triangle.v[0];
  //output<<"    2nd vertex ";
  output.width(10);
  output<< triangle.v[1];
  //output<<"    3rd vertex ";
  output.width(10);
  output<< triangle.v[2]<< endl;
}


ofstream& operator<< (ofstream& output, const CTriangle& triangle){
  output.width(10);
  output<< triangle.v[0];
  output.width(10);
  output<< triangle.v[1];
  output.width(10);
  output<< triangle.v[2]<< endl;
}

